@@ -4,7 +4,8 @@
 #    a drop-in replacement for IO::Socket::INET that encapsulates
 #    data passed over a network with SSL.
 #
-# Current Code Shepherd: Peter Behroozi, <behrooz at fas.harvard.edu>
+# Current Code Shepherd: Steffen Ullrich <steffen at genua.de>
+# Code Shepherd before: Peter Behroozi, <behrooz at fas.harvard.edu>
 #
 # The original version of this module was written by 
 # Marko Asplund, <marko.asplund at kronodoc.fi>, who drew from
@@ -40,7 +41,7 @@ use vars qw(@ISA $VERSION $DEBUG $SSL_ERROR $GLOBAL_CONTEXT_ARGS @EXPORT );
 BEGIN {
     # Declare @ISA, $VERSION, $GLOBAL_CONTEXT_ARGS
     @ISA = qw(IO::Socket::INET);
-    $VERSION = '1.0';
+    $VERSION = '1.06_1';
     $GLOBAL_CONTEXT_ARGS = {};
 
     #Make $DEBUG another name for $Net::SSLeay::trace
@@ -56,6 +57,8 @@ BEGIN {
 
 }
 
+sub CLONE_SKIP { 1 }
+
 # Export some stuff
 # inet4|inet6|debug will be handeled by myself, everything
 # else will be handeld the Exporter way
@@ -136,7 +139,7 @@ sub configure_SSL {
     %$arg_hash = ( %default_args, %$GLOBAL_CONTEXT_ARGS, %$arg_hash );
 
     #Avoid passing undef arguments to Net::SSLeay
-    !defined($arg_hash->{$_}) and ($arg_hash->{$_} = '') foreach (keys %$arg_hash);
+    defined($arg_hash->{$_}) or delete($arg_hash->{$_}) foreach (keys %$arg_hash);
 
     #Handle CA paths properly if no CA file is specified
     if ($arg_hash->{'SSL_ca_path'} ne '' and !(-f $arg_hash->{'SSL_ca_file'})) {
@@ -209,6 +212,7 @@ sub connect_SSL {
 	    	|| return $self->error("Failed to set SSL cipher list");
 	}
 
+	$arg_hash->{PeerAddr} || $self->_update_peer;
 	my $session = $ctx->session_cache( $arg_hash->{PeerAddr}, $arg_hash->{PeerPort} );
 	Net::SSLeay::set_session($ssl, $session) if ($session);
     }
@@ -243,6 +247,7 @@ sub connect_SSL {
     $ctx ||= ${*$self}{'_SSL_ctx'};
     if ( $ctx->has_session_cache ) {
 	my $arg_hash = ${*$self}{'_SSL_arguments'};
+	$arg_hash->{PeerAddr} || $self->_update_peer;
 	my ($addr,$port) = ( $arg_hash->{PeerAddr}, $arg_hash->{PeerPort} );
 	my $session = $ctx->session_cache( $addr,$port );
 	$ctx->session_cache( $addr,$port, Net::SSLeay::get1_session($ssl) ) if !$session;
@@ -253,6 +258,18 @@ sub connect_SSL {
     return $self;
 }
 
+# called if PeerAddr is not set in ${*$self}{'_SSL_arguments'}
+# this can be the case if start_SSL is called with a normal IO::Socket::INET
+# so that PeerAddr|PeerPort are not set from args
+sub _update_peer {
+    my $self = shift;
+    my $arg_hash = ${*$self}{'_SSL_arguments'};
+	eval {
+		my ($port,$addr) = sockaddr_in( getpeername( $self ));
+		$arg_hash->{PeerAddr} = inet_ntoa( $addr );
+		$arg_hash->{PeerPort} = $port;
+	}
+}
 
 #Call to accept occurs when a new client connects to a server using
 #IO::Socket::SSL
@@ -416,7 +433,7 @@ sub generic_write {
 # if all data are written
 sub write {
     my $self = shift;
-    return $self->generic_write( $self->blocking,@_ );
+    return $self->generic_write( scalar($self->blocking),@_ );
 }
 
 # contrary to write syswrite() returns already if only
@@ -713,6 +730,10 @@ sub set_default_context {
     $GLOBAL_CONTEXT_ARGS->{'SSL_reuse_ctx'} = shift;
 }
 
+sub set_default_session_cache {
+    $GLOBAL_CONTEXT_ARGS->{SSL_session_cache} = shift;
+}
+
 
 sub opened {
     my $self = shift;
@@ -748,6 +769,7 @@ sub recv   { croak("Use of recv() not implemented in IO::Socket::SSL; use read/s
 package IO::Socket::SSL::SSL_HANDLE;
 use strict;
 use vars qw($HAVE_WEAKREF);
+use Errno 'EBADF';
 
 BEGIN {
     local ($@, $SIG{__DIE__});
@@ -758,6 +780,8 @@ BEGIN {
     $HAVE_WEAKREF = $@ ? 0 : 1;
 }
 
+sub CLONE_SKIP { 1 }
+
 sub TIEHANDLE {
     my ($class, $handle) = @_;
     weaken($handle) if $HAVE_WEAKREF;
@@ -774,6 +798,9 @@ sub WRITE    { ${shift()}->syswrite (@_) }
 
 sub FILENO   { ${shift()}->fileno   (@_) }
 
+sub TELL     { $! = EBADF; return -1 }
+sub BINMODE  { return 0 }  # not perfect, but better than not implementing the method
+
 sub CLOSE {                          #<---- Do not change this function!
     my $ssl = ${$_[0]};
     local @_;
@@ -788,6 +815,8 @@ use strict;
 use constant SSL_MODE_ENABLE_PARTIAL_WRITE => 1;
 use constant SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER => 2;
 
+sub CLONE_SKIP { 1 }
+
 # Note that the final object will actually be a reference to the scalar
 # (C-style pointer) returned by Net::SSLeay::CTX_*_new() so that
 # it can be blessed.
@@ -910,13 +939,13 @@ sub new {
     Net::SSLeay::CTX_set_verify($ctx, $verify_mode, $verify_callback);
 
     $ctx_object = { context => $ctx };
-    if ($arg_hash->{'SSL_session_cache_size'}) {
-	if ($Net::SSLeay::VERSION < 1.26) {
-	    return IO::Socket::SSL->error("Session caches not supported for Net::SSLeay < v1.26");
-	} else {
-	    $ctx_object->{'session_cache'} =
-		IO::Socket::SSL::Session_Cache->new($arg_hash) || undef;
-	}
+    if ( my $cache = $arg_hash->{SSL_session_cache} ) {
+	# use predefined cache
+    	$ctx_object->{session_cache} = $cache
+    } elsif ( my $size = $arg_hash->{SSL_session_cache_size}) {
+	return IO::Socket::SSL->error("Session caches not supported for Net::SSLeay < v1.26")
+		if $Net::SSLeay::VERSION < 1.26;
+	$ctx_object->{session_cache} = IO::Socket::SSL::Session_Cache->new( $size );
     }
 
     return bless $ctx_object, $class;
@@ -925,19 +954,16 @@ sub new {
 
 sub session_cache {
     my $ctx = shift;
-    my $cache = $ctx->{'session_cache'};
-    return unless defined $cache;
-    my ($addr, $port) = (shift, shift);
+    my $cache = $ctx->{'session_cache'} || return;
+    my ($addr,$port,$session) = @_;
     my $key = "$addr:$port";
-    my $session = shift;
-
-    return (defined($session) ? $cache->add_session($key, $session)
-			      : $cache->get_session($key));
+    return defined($session) 
+    	? $cache->add_session($key, $session)
+	: $cache->get_session($key);
 }
 
 sub has_session_cache {
-    my $ctx = shift;
-    return (defined $ctx->{'session_cache'});
+    return defined shift->{session_cache};
 }
 
 
@@ -951,11 +977,12 @@ sub DESTROY {
 package IO::Socket::SSL::Session_Cache;
 use strict;
 
+sub CLONE_SKIP { 1 }
+
 sub new {
-    my ($class, $arg_hash) = @_;
-    my $cache = { _maxsize => $arg_hash->{'SSL_session_cache_size'}};
-    return unless ($cache->{_maxsize} > 0);
-    return bless $cache, $class;
+    my ($class, $size) = @_;
+    $size>0 or return;
+    return bless { _maxsize => $size }, $class;
 }
 
 
@@ -974,12 +1001,11 @@ sub get_session {
 
 sub add_session {
     my ($self, $key, $val) = @_;
-
     return if ($key eq '_maxsize' or $key eq '_head');
 
     if ((keys %$self) > $self->{'_maxsize'} + 1) {
 	my $last = $self->{'_head'}->{prev};
-	&Net::SSLeay::SESSION_free($last->{session});
+	Net::SSLeay::SESSION_free($last->{session});
 	delete($self->{$last->{key}});
 	$self->{'_head'}->{prev} = $self->{'_head'}->{prev}->{prev};
 	delete($self->{'_head'}) if ($self->{'_maxsize'} == 1);
@@ -1198,6 +1224,18 @@ The session cache size refers to the number of unique host/port pairs that can b
 stored at one time; the oldest sessions in the cache will be removed if new ones are
 added.  
 
+=item SSL_session_cache
+
+Specifies session cache object which should be used instead of creating a new.
+Overrules SSL_session_cache_size.
+This option is useful if you wan't to reuse the cache, but not the rest of
+the context.
+
+A session cache object can be created using 
+C<< IO::Socket::SSL::Session_Cache->new( cachesize ) >>.
+
+Use set_default_session_cache() to set a global cache object.
+
 =item SSL_error_trap
 
 When using the accept() or connect() methods, it may be the case that the
@@ -1312,6 +1350,16 @@ be either an IO::Socket::SSL object or an IO::Socket::SSL::SSL_Context object.
 the SSL_reuse_ctx option of new() for more details.  Note that this sets the default
 context globally, so use with caution (esp. in mod_perl scripts).
 
+=item B<IO::Socket::SSL::set_default_session_cache(...)>
+
+You may use this to make IO::Socket::SSL automatically re-use a given session cache
+(unless specifically overridden in a call to new()).  It accepts one argument, which should
+be an IO::Socket::SSL::Session_Cache object or similar (e.g something which implements
+get_session and add_session like IO::Socket::SSL::Session_Cache does).
+See the SSL_session_cache option of new() for more details.  Note that this sets the default
+cache globally, so use with caution.
+
+
 =back
 
 The following methods are unsupported (not to mention futile!) and IO::Socket::SSL
@@ -1412,6 +1460,9 @@ This is because IO::Socket::SSL is based on Net::SSLeay which
 uses a global object to access some of the API of openssl
 and is therefore not threadsafe.
 
+IO::Socket::SSL does not work together with Storable::fd_retrieve/fd_store.
+See BUGS file for more information and how to work around the problem.
+
 =head1 LIMITATIONS
 
 IO::Socket::SSL uses Net::SSLeay as the shiny interface to OpenSSL, which is
